#include "constraints.h"

namespace MaskTools { namespace Constraints { 

Modulo toModulo(int n)
{
   if ( !(n & MODULO_32) )
      return MODULO_32;
   else if ( !(n & MODULO_16) )
      return MODULO_16;
   else if ( !(n & MODULO_8) )
      return MODULO_8;
   else if ( !(n & MODULO_4) )
      return MODULO_4;
   else if ( !(n & MODULO_2) )
      return MODULO_2;
   else
      return MODULO_NONE;
}

Alignment toAlignment(int n)
{
   if ( !(n & ALIGNMENT_32) )
      return ALIGNMENT_32;
   else if ( !(n & ALIGNMENT_16) )
      return ALIGNMENT_16;
   else if ( !(n & ALIGNMENT_8) )
      return ALIGNMENT_8;
   else if ( !(n & ALIGNMENT_4) )
      return ALIGNMENT_4;
   else if ( !(n & ALIGNMENT_2) )
      return ALIGNMENT_2;
   else
      return ALIGNMENT_NONE;
}

Constraint::Constraint() : flags(CPU_NONE) { }

Constraint::Constraint(CpuFlags flags, Modulo moduloX, Modulo moduloY, Alignment alignment) :
flags(flags), moduloX(moduloX), moduloY(moduloY), alignment(alignment) { }

Constraint::Constraint(CpuFlags flags, int nWidth, int nHeight, const Byte *pAddress, int nPitch) :
flags(flags), moduloX(toModulo(nWidth)), moduloY(toModulo(nHeight)), alignment(toAlignment(int(pAddress) | nPitch)) { }

Constraint::Constraint(CpuFlags flags, int nWidth, int nHeight, int nAddress, int nPitch) :
flags(flags), moduloX(toModulo(nWidth)), moduloY(toModulo(nHeight)), alignment(toAlignment(nAddress | nPitch)) { }

bool Constraint::respect(const Constraint &constraint) const
{
   return ((flags & constraint.flags) == flags) && (moduloX <= constraint.moduloX) && 
          (moduloY <= constraint.moduloY) && (alignment <= constraint.alignment);
}

} }
