#ifndef __Mt_Symbol_H__
#define __Mt_Symbol_H__

#include "../../common/common.h"

namespace MaskTools { namespace Parser {

class Symbol {
public:
   typedef enum {
      NUMBER,
      OPERATOR,
      FUNCTION,
      TERNARY,
      VARIABLE,

      UNDEFINED

   } Type;

public:

   Type type;
   String value;
   int nParameter;

private:

   double dValue;
   typedef double (*Process)(double x, double y, double z);
private:
   Process process;
public:

   Symbol();
   Symbol(String value, Type type, int nParameter, Process process);
   Symbol(String value, double dValue, Type type, int nParameter, Process process);

   void setValue(double dValue);
   double getValue(double x, double y, double z) const;

   static Symbol Addition;
   static Symbol Multiplication;
   static Symbol Division;
   static Symbol Substraction;
   static Symbol Power;
   static Symbol Modulo;
   static Symbol Interrogation;
   static Symbol Equal;
   static Symbol NotEqual;
   static Symbol Inferior;
   static Symbol InferiorStrict;
   static Symbol Superior;
   static Symbol SuperiorStrict;
   static Symbol And;
   static Symbol Or;
   static Symbol AndNot;
   static Symbol Xor;
   static Symbol Pi;
   static Symbol X;
   static Symbol Y;
   static Symbol Cos;
   static Symbol Sin;
   static Symbol Tan;
   static Symbol Log;
   static Symbol Abs;
   static Symbol Exp;
   static Symbol Acos;
   static Symbol Atan;
   static Symbol Asin;
};

class Context : private std::list<double> {

   std::list<Symbol> expression;

   double popb();

public:
   
   Context(const std::list<Symbol> &expression);
   Context &setVariable(const String &name, double value);

   double compute();
   bool check();
   Byte compute_byte() { double f = compute(); return f < 0.0f ? 0 : (f > 255.0f ? 255 : Byte(f + 0.5f)); }
};

} }

#endif
