#include "filter.h"
#include "../../common/parser/parser.h"

using namespace MaskTools;
using namespace MaskTools::Picture;
using namespace MaskTools::Parameters;
using namespace MaskTools::Constraints;
using namespace MaskTools::Parser;

namespace MaskTools { namespace Filters { namespace Convolution {

void Filter::ProcessY(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(),
                                      horizontal, vertical, nHorizontal, nVertical, nWidth, nHeight);
}

void Filter::ProcessU(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(),
                                      horizontal, vertical, nHorizontal, nVertical, nWidthUV, nHeightUV);
}

void Filter::ProcessV(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(),
                                      horizontal, vertical, nHorizontal, nVertical, nWidthUV, nHeightUV);
}

Filter::Filter(const Parameters::Parameters &parameters) : MaskTools::Filter(parameters)
{
   i_vertical = i_horizontal = NULL;
   f_vertical = f_horizontal = NULL;
   vertical = horizontal = NULL;
   std::list<Symbol> hcoeffs = Parser::getDefaultParser().parse(parameters["horizontal"].toString(), " ").getExpression();
   std::list<Symbol> vcoeffs = Parser::getDefaultParser().parse(parameters["vertical"].toString(), " ").getExpression();
   nHorizontal = hcoeffs.size();
   nVertical = vcoeffs.size();
   
   /* search for float values */
   bool isFloat = false;
   isFloat |= parameters["horizontal"].toString().find(".", 0) != String::npos;
   isFloat |= parameters["vertical"].toString().find(".", 0) != String::npos;

   /* create the two arrays */
   if ( isFloat )
   {
      f_vertical = new float[nVertical];
      f_horizontal = new float[nHorizontal];
      vertical = f_vertical;
      horizontal = f_horizontal;
   }
   else
   {
      i_vertical = new int[nVertical];
      i_horizontal = new int[nHorizontal];
      vertical = i_vertical;
      horizontal = i_horizontal;
   }

   /* fill them */
   for ( int i = 0; i < nHorizontal; i++ )
   {
      if ( isFloat )
         f_horizontal[i] = float(hcoeffs.front().getValue(0,0,0));
      else
         i_horizontal[i] = int(hcoeffs.front().getValue(0,0,0) + 0.5f);

      hcoeffs.pop_front();
   }
   for ( int i = 0; i < nVertical; i++ )
   {
      if ( isFloat )
         f_vertical[i] = float(vcoeffs.front().getValue(0,0,0));
      else
         i_vertical[i] = int(vcoeffs.front().getValue(0,0,0) + 0.5f);

      vcoeffs.pop_front();
   }

   /* adds the processor */
   if ( isFloat )
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(convolution_f_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
   else
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(convolution_i_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
}

Filter::~Filter()
{
}

String Filter::getName() { return "mt_convolution"; }

Signature Filter::getSignature()
{
   Signature signature;
   signature.push_back(Parameter(CLIP, ""));
   signature.push_back(Parameter(String("1 1 1"), "horizontal"));
   signature.push_back(Parameter(String("1 1 1"), "vertical"));

   return addDefaultCommons(signature);
}

Clip::Inputs Filter::inputClips = Clip::OneFrame();

} } }