#include "filter.h"

using namespace MaskTools;
using namespace MaskTools::Picture;
using namespace MaskTools::Parameters;
using namespace MaskTools::Constraints;

namespace MaskTools { namespace Filters { namespace Logic {

void Filter::ProcessY(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), nWidth, nHeight);
}

void Filter::ProcessU(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), nWidthUV, nHeightUV);
}

void Filter::ProcessV(PlaneProcessor *processor, Plane &dst, CPlane &src1,
                      CPlane &src2, CPlane &src3)
{
   static_cast<Processor*>(processor)(dst.plane(), dst.pitch(), src1.plane(), src1.pitch(), nWidthUV, nHeightUV);
}

Filter::Filter(const Parameters::Parameters &parameters) : MaskTools::Filter(parameters)
{
   /* add the processors */
   if ( parameters["mode"].toString() == "and" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(and_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_and8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_andn_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_and8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 3));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_and8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 4));
   }
   else if ( parameters["mode"].toString() == "andn" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(andn_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_andn8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_andnn_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_andn8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 3));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_andn8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 4));
   }
   else if ( parameters["mode"].toString() == "or" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(or_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_or8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_orn_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_or8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 3));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_or8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 4));
   }
   else if ( parameters["mode"].toString() == "xor" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(xor_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_xor8_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_xorn_mmx), Constraint(CPU_MMX, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_xor8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 3));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_xor8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 4));
   }
   else if ( parameters["mode"].toString() == "min" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(min_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_min8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_minn_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_min8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 3));
   }
   else if ( parameters["mode"].toString() == "max" )
   {
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(max_c), Constraint(CPU_NONE, 1, 1, 1, 1), 0));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_max8_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 1));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_maxn_isse), Constraint(CPU_ISSE, 8, 1, 1, 1), 2));
      addProcessor(Constraints::Processor<PlaneProcessor>(PPlaneProcessor(Logic_max8_3dnow), Constraint(CPU_3DNOW, 8, 1, 1, 1), 3));
   }
   else
   {
      error = "logic : unknown mode. Available modes are 'and', 'or', 'xor', 'andn', 'min' and 'max'";
      return;
   }
}

String Filter::getName() { return "mt_logic"; }

Signature Filter::getSignature()
{
   Signature signature;
   signature.push_back(Parameter(CLIP, ""));
   signature.push_back(Parameter(CLIP, ""));
   signature.push_back(Parameter(String("and"), "mode"));

   return addDefaultCommons(signature);
}

Clip::Inputs Filter::inputClips = Clip::IPTwoFrames();

} } }