%INCLUDE "common.asm"

SECTION .rodata:
ALIGN 16
MT_BYTE_7F : TIMES 8 DB 0x7F

;=============================================================================
; Code
;=============================================================================

SECTION .text:

   mt_mangle Edge, sobel8_mmx
   mt_mangle Edge, roberts8_mmx
   mt_mangle Edge, laplace8_mmx
   mt_mangle Edge, convolution8_mmx
   
%MACRO MT_SOBEL_ROLLED_4_INSIDE  14
   %6                      movq  , %11, esi + %1, ebx, mm6, mm7
   %8                      movq  , %12, esi + %1, ebx, mm6, mm7
   %5                      movq  , %13, esi + %1, ebx, mm6, mm7
   %3                      movq  , %14, esi + %1, ebx, mm6, mm7
   
   %10                     %11   , mm5
   %10                     %12   , mm5
   %10                     %13   , mm5
   %10                     %14   , mm5
   
   paddw                   %11   , %12
   paddw                   %13   , %14
   movq                    %14   , %11
   psubusw                 %11   , %13
   psubusw                 %13   , %14
   paddw                   %11   , %13
   psrlw                   %11   , 1
%ENDMACRO   

%MACRO MT_ROBERTS_ROLLED_4_INSIDE  14
   movq                    %11   , [esi + ebx + %1]
   %6                      movq  , %12, esi + %1, ebx, mm6, mm7
   %8                      movq  , %13, esi + %1, ebx, mm6, mm7
   %10                     %11   , mm5
   %10                     %12   , mm5
   %10                     %13   , mm5
   psllw                   %11   , 1
   paddw                   %12   , %13
   movq                    %14   , %11
   psubusw                 %11   , %12
   psubusw                 %12   , %14
   paddw                   %11   , %12
   psrlw                   %11   , 1
%ENDMACRO

%MACRO MT_LAPLACE_ROLLED_4_INSIDE  14
   %2                      movq  , %11, esi + %1, ebx, mm6, mm7
   %3                      movq  , %12, esi + %1, ebx, mm6, mm7
   %4                      movq  , %13, esi + %1, ebx, mm6, mm7
   %5                      movq  , %14, esi + %1, ebx, mm6, mm7
   %10                     %11   , mm5
   %10                     %12   , mm5
   %10                     %13   , mm5
   %10                     %14   , mm5
   paddw                   %11   , %12
   paddw                   %13   , %14
   paddw                   %11   , %13
   %6                      movq  , %12, esi + %1, ebx, mm6, mm7
   %7                      movq  , %13, esi + %1, ebx, mm6, mm7
   %8                      movq  , %14, esi + %1, ebx, mm6, mm7
   %10                     %12   , mm5
   %10                     %13   , mm5
   %10                     %14   , mm5
   paddw                   %12   , %13
   paddw                   %11   , %14
   paddw                   %11   , %12
   %9                      movq  , %13, esi + %1, ebx, mm6, mm7
   movq                    %12   , [esi + ebx + %1]
   %10                     %13   , mm5
   %10                     %12   , mm5
   paddw                   %11   , %13
   psllw                   %12   , 3
   movq                    %13   , %11
   psubusw                 %11   , %12
   psubusw                 %12   , %13
   paddw                   %11   , %12
   psrlw                   %11   , 3
%ENDMACRO

%MACRO MT_CONVOLUTION_ROLLED_4_INSIDE  14
   %2                      movq  , %11, esi + %1, ebx, mm6, mm7
   %3                      movq  , %12, esi + %1, ebx, mm6, mm7
   %4                      movq  , %13, esi + %1, ebx, mm6, mm7
   %5                      movq  , %14, esi + %1, ebx, mm6, mm7
   %10                     %11   , mm5
   %10                     %12   , mm5
   %10                     %13   , mm5
   %10                     %14   , mm5
   pmullw                  %11   , [esp + .nTopLeft]
   pmullw                  %12   , [esp + .nTop]
   pmullw                  %13   , [esp + .nTopRight]
   pmullw                  %14   , [esp + .nLeft]
   paddw                   %11   , %12
   paddw                   %13   , %14
   paddw                   %11   , %13
   %6                      movq  , %12, esi + %1, ebx, mm6, mm7
   %7                      movq  , %13, esi + %1, ebx, mm6, mm7
   %8                      movq  , %14, esi + %1, ebx, mm6, mm7
   %10                     %12   , mm5
   %10                     %13   , mm5
   %10                     %14   , mm5
   pmullw                  %12   , [esp + .nRight]
   pmullw                  %13   , [esp + .nBottomLeft]
   pmullw                  %14   , [esp + .nBottom]
   paddw                   %12   , %13
   paddw                   %11   , %14
   paddw                   %11   , %12
   %9                      movq  , %13, esi + %1, ebx, mm6, mm7
   movq                    %12   , [esi + ebx + %1]
   %10                     %13   , mm5
   %10                     %12   , mm5
   pmullw                  %13   , [esp + .nBottomRight]
   pmullw                  %12   , [esp + .nCenter]
   paddw                   %11   , %13
   
   movq                    %13   , %11
   psubusw                 %11   , %12
   psubusw                 %12   , %13
   paddw                   %11   , %12
   
   psrlw                   %11   , [esp + .nDivisor]
%ENDMACRO


%MACRO MT_EDGE_THRESHOLDING 1
   packuswb                mm0   , mm1                            ; abs values
   movq                    mm1   , mm0
   psubb                   mm1   , [MT_BYTE_7F]
   movq                    mm3   , mm1
   pcmpgtb                 mm1   , [esp + .nLow8]
   pcmpgtb                 mm3   , [esp + .nHigh8]
   pand                    mm0   , mm1
   por                     mm0   , mm3
   
   movq                    [edi + %1], mm0
%ENDMACRO   

%MACRO MT_SOBEL_ROLLED_INSIDE    9
   MT_SOBEL_ROLLED_4_INSIDE      %1, %2, %3, %4, %5, %6, %7, %8, %9, punpcklbw, mm0, mm1, mm2, mm3
   MT_SOBEL_ROLLED_4_INSIDE      %1, %2, %3, %4, %5, %6, %7, %8, %9, punpckhbw, mm1, mm2, mm3, mm4
   MT_EDGE_THRESHOLDING          %1
%ENDMACRO

%MACRO MT_ROBERTS_ROLLED_INSIDE  9
   MT_ROBERTS_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpcklbw, mm0, mm1, mm2, mm3
   MT_ROBERTS_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpckhbw, mm1, mm2, mm3, mm4
   MT_EDGE_THRESHOLDING          %1
%ENDMACRO

%MACRO MT_LAPLACE_ROLLED_INSIDE  9
   MT_LAPLACE_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpcklbw, mm0, mm1, mm2, mm3
   MT_LAPLACE_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpckhbw, mm1, mm2, mm3, mm4
   MT_EDGE_THRESHOLDING          %1
%ENDMACRO

%MACRO MT_CONVOLUTION_ROLLED_INSIDE  9
   MT_CONVOLUTION_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpcklbw, mm0, mm1, mm2, mm3
   MT_CONVOLUTION_ROLLED_4_INSIDE    %1, %2, %3, %4, %5, %6, %7, %8, %9, punpckhbw, mm1, mm2, mm3, mm4
   MT_EDGE_THRESHOLDING             %1
%ENDMACRO

%MACRO MT_EDGE_ROLLED            9
   MT_ %+ %1 %+ _ROLLED_INSIDE   0 , %2, %3, %4, %5, %6, %7, %8, %9
%ENDMACRO

%MACRO MT_SOBEL_ROLLED           8
   MT_EDGE_ROLLED                SOBEL, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_ROBERTS_ROLLED         8
   MT_EDGE_ROLLED                ROBERTS, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_LAPLACE_ROLLED         8
   MT_EDGE_ROLLED                LAPLACE, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_CONVOLUTION_ROLLED     8
   MT_EDGE_ROLLED                CONVOLUTION, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_EDGE_UNROLLED          9
   MT_ %+ %1 %+ _ROLLED_INSIDE   0 , %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   8 , %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   16, %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   24, %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   32, %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   40, %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   48, %2, %3, %4, %5, %6, %7, %8, %9
   MT_ %+ %1 %+ _ROLLED_INSIDE   56, %2, %3, %4, %5, %6, %7, %8, %9
%ENDMACRO

%MACRO MT_SOBEL_UNROLLED         8
   MT_EDGE_UNROLLED              SOBEL, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_ROBERTS_UNROLLED       8
   MT_EDGE_UNROLLED              ROBERTS, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_LAPLACE_UNROLLED       8
   MT_EDGE_UNROLLED              LAPLACE, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_CONVOLUTION_UNROLLED       8
   MT_EDGE_UNROLLED              CONVOLUTION, %1, %2, %3, %4, %5, %6, %7, %8
%ENDMACRO

%MACRO MT_EDGE_NEXT_LINE         0
   add                           esi, [esp + .nSrcPad]
   add                           edi, [esp + .nDstPad]
%ENDMACRO

%MACRO MT_EDGE_NEXT_8_PIXELS     0
   add                           esi, 8
   add                           edi, 8
%ENDMACRO

%MACRO MT_EDGE_NEXT_64_PIXELS    0
   add                           esi, 64
   add                           edi, 64
%ENDMACRO

%MACRO MT_EDGE_FUNCTION          2
   MT_SET_STACK            9, esi, edi, ebx, ebp
   .pDst                   PARAM(1)
   .nDstPitch              PARAM(2)
   .pSrc                   PARAM(3)
   .nSrcPitch              PARAM(4)
   ; matrix
   .nLowThreshold          PARAM(6)
   .nHighThreshold         PARAM(7)
   .nWidth                 PARAM(8)
   .nHeight                PARAM(9)
   .nLow8                  PSTACK(0)
   .nHigh8                 PSTACK(2)
   .nTailWidth             PSTACK(4)
   .nUnrolledWidth         PSTACK(5)
   .nSrcPad                PSTACK(6)
   .nDstPad                PSTACK(7)
   .nHeightCount           PSTACK(8)
   
   pxor                    mm5, mm5
   
   movd                    mm6, [esp + .nLowThreshold]
   movd                    mm7, [esp + .nHighThreshold]
   punpcklbw               mm6, mm6
   punpcklbw               mm7, mm7
   punpcklwd               mm6, mm6
   punpcklwd               mm7, mm7
   punpckldq               mm6, mm6
   punpckldq               mm7, mm7
   psubb                   mm6, [MT_BYTE_7F]
   psubb                   mm7, [MT_BYTE_7F]
   movq                    [esp + .nLow8], mm6
   movq                    [esp + .nHigh8], mm7
   
   mov                     edi, [esp + .pDst]
   mov                     esi, [esp + .pSrc]
   mov                     ebx, [esp + .nSrcPitch]
   mov                     ebp, [esp + .nDstPitch]
   sub                     esi, ebx                         ; positive offset needed for pitches
   
   mov                     ecx, [esp + .nWidth]
   sub                     ecx, ebx
   neg                     ecx
   mov                     [esp + .nSrcPad], ecx            ; padding src
   mov                     ecx, [esp + .nWidth]
   sub                     ecx, ebp
   neg                     ecx
   mov                     [esp + .nDstPad], ecx            ; padding dst
   
   MT_PROCESS_CONVOLUTION  %1, %2, MT_EDGE_NEXT_LINE, MT_EDGE_NEXT_8_PIXELS, MT_EDGE_NEXT_64_PIXELS
   
   MT_REL_STACK
%ENDMACRO

sobel8_mmx :
   MT_EDGE_FUNCTION        MT_SOBEL_UNROLLED, MT_SOBEL_ROLLED
   
roberts8_mmx :
   MT_EDGE_FUNCTION        MT_ROBERTS_UNROLLED, MT_ROBERTS_ROLLED
   
laplace8_mmx :
   MT_EDGE_FUNCTION        MT_LAPLACE_UNROLLED, MT_LAPLACE_ROLLED
   
%MACRO MT_LOAD_WORD        2
   movd                    mm6, [eax + %1 * 2]
   punpcklwd               mm6, mm6
   punpckldq               mm6, mm6
   movq                    [esp + %2], mm6
%ENDMACRO
   
%MACRO MT_CONVOLUTION_FUNCTION 0
   MT_SET_STACK            29, esi, edi, ebx, ebp
   .pDst                   PARAM(1)
   .nDstPitch              PARAM(2)
   .pSrc                   PARAM(3)
   .nSrcPitch              PARAM(4)
   .matrix                 PARAM(5)
   .nLowThreshold          PARAM(6)
   .nHighThreshold         PARAM(7)
   .nWidth                 PARAM(8)
   .nHeight                PARAM(9)
   .nLow8                  PSTACK(0)
   .nHigh8                 PSTACK(2)
   .nTopLeft               PSTACK(4)
   .nTop                   PSTACK(6)
   .nTopRight              PSTACK(8)
   .nLeft                  PSTACK(10)
   .nRight                 PSTACK(12)
   .nBottomLeft            PSTACK(14)
   .nBottom                PSTACK(16)
   .nBottomRight           PSTACK(18)
   .nCenter                PSTACK(20)
   .nDivisor               PSTACK(22)
   .nTailWidth             PSTACK(24)
   .nUnrolledWidth         PSTACK(25)
   .nSrcPad                PSTACK(26)
   .nDstPad                PSTACK(27)
   .nHeightCount           PSTACK(28)
   
   pxor                    mm5, mm5
   
   movd                    mm6, [esp + .nLowThreshold]
   movd                    mm7, [esp + .nHighThreshold]
   punpcklbw               mm6, mm6
   punpcklbw               mm7, mm7
   punpcklwd               mm6, mm6
   punpcklwd               mm7, mm7
   punpckldq               mm6, mm6
   punpckldq               mm7, mm7
   psubb                   mm6, [MT_BYTE_7F]
   psubb                   mm7, [MT_BYTE_7F]
   movq                    [esp + .nLow8], mm6
   movq                    [esp + .nHigh8], mm7
       
   mov                     eax, [esp + .matrix]

   MT_LOAD_WORD            0, .nTopLeft
   MT_LOAD_WORD            1, .nTop
   MT_LOAD_WORD            2, .nTopRight
   MT_LOAD_WORD            3, .nLeft
   MT_LOAD_WORD            5, .nRight
   MT_LOAD_WORD            6, .nBottomLeft
   MT_LOAD_WORD            7, .nBottom
   MT_LOAD_WORD            8, .nBottomRight

   movd                    mm6, [eax + 8]
   punpcklwd               mm6, mm6
   punpckldq               mm6, mm6
   pxor                    mm7, mm7
   psubw                   mm7, mm6
   movq                    [esp + .nCenter], mm7
   
   ;MT_LOAD_WORD            4, .nCenter
   
   mov                     cx, [eax + 18]
   xor                     eax, eax
   mov                     [esp + .nDivisor + 4], eax
.loop_div
   inc                     eax
   shr                     cx, 1
   jnz                     .loop_div
   dec                     eax
   mov                     [esp + .nDivisor], eax
   
   mov                     edi, [esp + .pDst]
   mov                     esi, [esp + .pSrc]
   mov                     ebx, [esp + .nSrcPitch]
   mov                     ebp, [esp + .nDstPitch]
   sub                     esi, ebx                         ; positive offset needed for pitches
   
   mov                     ecx, [esp + .nWidth]
   sub                     ecx, ebx
   neg                     ecx
   mov                     [esp + .nSrcPad], ecx            ; padding src
   mov                     ecx, [esp + .nWidth]
   sub                     ecx, ebp
   neg                     ecx
   mov                     [esp + .nDstPad], ecx            ; padding dst
   
   MT_PROCESS_CONVOLUTION  MT_CONVOLUTION_UNROLLED, MT_CONVOLUTION_ROLLED, MT_EDGE_NEXT_LINE, MT_EDGE_NEXT_8_PIXELS, MT_EDGE_NEXT_64_PIXELS
   
   MT_REL_STACK
%ENDMACRO
   
convolution8_mmx :
   MT_CONVOLUTION_FUNCTION