#include "filter.h"
#include "../../../filters/mask/functions.h"

using namespace MaskTools;

namespace MaskTools { namespace Filters { namespace Mask { namespace Edge {

static inline Byte convolution(Byte a11, Byte a21, Byte a31, Byte a12, Byte a22, Byte a32, Byte a13, Byte a23, Byte a33, const Word matrix[10], int nLowThreshold, int nHighThreshold)
{
   return applyThresholds(abs((a11 * matrix[0] + a21 * matrix[1] + a31 * matrix[2] + 
                               a12 * matrix[3] + a22 * matrix[4] + a32 * matrix[5] +
                               a13 * matrix[6] + a23 * matrix[7] + a33 * matrix[8]) / matrix[9]), nLowThreshold, nHighThreshold);
}

static inline Byte sobel(Byte a11, Byte a21, Byte a31, Byte a12, Byte a22, Byte a32, Byte a13, Byte a23, Byte a33, const Word matrix[10], int nLowThreshold, int nHighThreshold)
{
   return applyThresholds(abs( (int)a32 + a23 - a12 - a21 ) >> 1, nLowThreshold, nHighThreshold);
}

static inline Byte roberts(Byte a11, Byte a21, Byte a31, Byte a12, Byte a22, Byte a32, Byte a13, Byte a23, Byte a33, const Word matrix[10], int nLowThreshold, int nHighThreshold)
{
   return applyThresholds(abs( ((int)a22 << 1) - a32 - a23 ) >> 1, nLowThreshold, nHighThreshold);
}

static inline Byte laplace(Byte a11, Byte a21, Byte a31, Byte a12, Byte a22, Byte a32, Byte a13, Byte a23, Byte a33, const Word matrix[10], int nLowThreshold, int nHighThreshold)
{
   return applyThresholds(abs( ((int)a22 << 3) - a32 - a23 - a11 - a21 - a31 - a12 - a13 - a33 ) >> 3, nLowThreshold, nHighThreshold);
}

class Thresholds {
   Byte nMinThreshold, nMaxThreshold;
public:
   Thresholds(Byte nMinThreshold, Byte nMaxThreshold) :
   nMinThreshold(nMinThreshold), nMaxThreshold(nMaxThreshold)
   {
   }

   int minpitch() const { return 0; }
   int maxpitch() const { return 0; }
   void nextRow() { }
   Byte min(int x) const { return nMinThreshold; }
   Byte max(int x) const { return nMaxThreshold; }
};

template<Filters::Mask::Operator op>
void mask_t(Byte *pDst, int nDstPitch, const Byte *pSrc, int nSrcPitch, const Word matrix[10], int nLowThreshold, int nHighThreshold, int nWidth, int nHeight)
{
   Filters::Mask::generic_c<op, Thresholds>(pDst, nDstPitch, pSrc, nSrcPitch, Thresholds(nLowThreshold, nHighThreshold), matrix, nWidth, nHeight);
}

extern "C" Processor Edge_sobel8_mmx;
extern "C" Processor Edge_roberts8_mmx;
extern "C" Processor Edge_laplace8_mmx;
extern "C" Processor Edge_convolution8_mmx;

Processor *convolution_c = &mask_t<convolution>;
Processor *convolution8_mmx = &Edge_convolution8_mmx;

Processor *sobel_c = &mask_t<sobel>;
Processor *sobel8_mmx = &Edge_sobel8_mmx;

Processor *roberts_c = &mask_t<roberts>;
Processor *roberts8_mmx = &Edge_roberts8_mmx;

Processor *laplace_c = &mask_t<laplace>;
Processor *laplace8_mmx = &Edge_laplace8_mmx;

} } } }